###################################################################
### MP Fénelon Sainte-Marie --- Informatique
### Séance 16 --- Graphes
###################################################################

import numpy as np
import matplotlib.pyplot as plt
import numpy.random as rd

plt.clf()

n = 7
graphe = np.zeros((n,n))

graphe[0,2] =  5 # A -> C
graphe[0,5] = 35 # A -> F
graphe[1,2] =  9 # B -> C
graphe[1,4] = 15 #´B -> E
graphe[2,4] = 10 # C -> E
graphe[2,5] =  8 # C -> F
graphe[3,0] =  3 # D -> A
graphe[3,1] = 12 # D -> B
graphe[4,6] = 14 # E -> G
graphe[5,6] = 13 # F -> G
#print("graphe :")
#print(graphe)

n = 100
print("nombre de sommets n =",n)
"""
graphe = rd.randint(0,4,(n,n)) # graphe aléatoire
for i in range(n):
    graphe[i,i] = 0
    for j in range(n):
        graphe[i,j] = graphe[i,j] * rd.randint(2)
"""

p = 0.995
graphe = rd.randint(0,10,(n,n)) # graphe aléatoire
for i in range(n):
    graphe[i,i] = 0
    for j in range(n):
        if rd.random()<p:
            graphe[i,j] = 0
# ajout d'un chemin circulaire afin que le graphe soit connexe par arcs
for i in range(n-1):
    if graphe[i,i+1]==0:
        graphe[i,i+1] = 50
if graphe[n-1,0]==0:
    graphe[n-1,0] = 50

#######################################################

def adjacent(g,n,s):
    # dans le graphe g à n sommets,
    # recherche des sommets s' tels que s -> s'
    # on renvoie la liste des sommets successeurs de s
    
    resu = []
    for j in range(n):
        if g[s,j]>0:
            resu = resu + [j]
    return resu
    
#print(adjacent(graphe,n,2))

#######################################################

def represente(g,n):
    
    for k in range(n):
        xk = np.cos(2*k*np.pi/n)
        yk = np.sin(2*k*np.pi/n)
                
        plt.plot(xk,yk,'ok')
        plt.text(1.1*xk,1.1*yk,r''+str(k))
        p = 0.80
        q = 1 - p
    for i in range(n):
        for j in range(n):
            if g[i,j] !=0:
                xi = np.cos(2*i*np.pi/n)
                xj = np.cos(2*j*np.pi/n)
                yi = np.sin(2*i*np.pi/n)
                yj = np.sin(2*j*np.pi/n)
                
                plt.arrow(xi, yi, 0.95*(xj-xi), 0.95*(yj-yi),hold=None,fc='b', ec='b')
                # on affiche le poids à proximité de la source
                plt.text(p*xi+q*xj,p*yi+q*yj,r''+str(g[i,j]))
                
    plt.axis([-1.2,1.2,-1.2,1.2])

#######################################################

def represente_chemin(g,n,L):
    # la liste L contient les numéros des sommets à relier
    
    X = [np.cos(2*i*np.pi/n) for i in L]
    Y = [np.sin(2*i*np.pi/n) for i in L]
    plt.plot(X,Y,'r',linewidth=3)

#represente_chemin(graphe,n,[3,0,2,5,6])

#######################################################

def nonfixes(liste,n):
    # renvoie une liste (éventuellement vide) contenant les indices des
    # sommets non encore fixés, encodant un parcours de liste
    resu = []
    for i in range(n):
        if liste[i][0]==False:
            resu = resu+[i]
    return resu

#######################################################

# Choisissons un sommet initial

s_deb = 3

# On initialise le graphe muni d'un sommet initial s
# en clair on construit une liste de n couples, chacun
# correspondant à un sommet :
# [booléen "ce sommet est fixé", distance au sommet initial, prédécesseur]

L = [[False,9999,-1] for i in range(n)]
L[s_deb][0] = True   # le sommet initial est fixé
L[s_deb][1] = 0      # le poids du sommet initial est nul

#print(L)

# le sommet courant

courant = s_deb
nombreFixes = 1

#a_visiter = [i for i in range(n)]
#print("avant il reste à visiter:",a_visiter)
#a_visiter.pop(courant) # on enlève l'élément de départ des sommets à visiter

#print("après il reste à visiter:",nonfixes(L,n))

while nombreFixes<n:
    #print("########## sommet courant :",courant,"##########")
    
    a_visiter = nonfixes(L,n)
    
    # pour tous les sommets à visiter, reliés au sommet courant :
    # - on indique leur prédécesseur : le sommet courant,
    # - on met à jour leur poids (si besoin)
        
    for i in a_visiter:
        
        if graphe[courant,i]>0:  # si courant->i alors on examine
            ancien_poids = L[i][1]  # ancien poids
            nouveau_poids = L[courant][1] + graphe[courant,i] # nouveau poids
            # si nouveau poids < ancien poids alors on met à jour
            if nouveau_poids<ancien_poids:
                L[i][2] = courant  # le prédécesseur de i est courant
                L[i][1] = nouveau_poids # mise à jour du poids
    
    # parmi les sommets à visiter, on recherche celui de poids minimal
    
    indice_mini = a_visiter[0]
    
    for i in a_visiter:
        if L[i][1]<L[indice_mini][1]:
            indice_mini = i
    
    # quand on a trouvé un sommet à visiter de plus petit poids,
    # on le marque comme "fixé"
    
    courant = indice_mini
    L[courant][0] = True
    
    nombreFixes = nombreFixes + 1

# recherche du chemin le plus court
s_arr = 6
chemin = [s_arr]
predecesseur = s_arr
while predecesseur!=s_deb:
    predecesseur = L[predecesseur][2]
    chemin.append(predecesseur)
l = len(chemin)
bon = [chemin[l-1-i] for i in range(l)]
print("le chemin le plus court de",s_deb,"vers",s_arr,"est : ",bon)
print("sa longueur est",l-1,"et son poids",L[s_arr][1])

represente(graphe,n)
#represente_chemin(graphe,n,bon)
